# -*- coding: utf-8 -*-
"""
Testing reader.py


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

from pathlib import Path
from tdvisu.reader import TwReader


def test_reader():
    """Read a file containing graph edges. Check stored edges and adjacency
    as well as the number of vertices and number of edges."""

    twfile = Path(__file__).parent / 'grda16.tw'

    expected_edges = {(1, 2), (2, 1), (2, 3), (3, 2), (3, 4), (3, 5),
                      (4, 3), (4, 5), (4, 6), (5, 3), (5, 4), (6, 4),
                      (6, 7), (6, 15), (7, 6), (7, 8), (7, 14), (8, 7),
                      (8, 9), (9, 8), (9, 10), (9, 11), (10, 9),
                      (11, 9), (11, 12), (11, 14), (12, 11), (12, 13),
                      (12, 14), (13, 12), (14, 7), (14, 11), (14, 12),
                      (15, 6), (15, 16), (16, 15)}

    expected_adj = {1: {2}, 2: {1, 3}, 3: {2, 4, 5}, 4: {3, 5, 6},
                    5: {3, 4}, 6: {4, 7, 15}, 7: {6, 8, 14}, 8: {7, 9},
                    9: {8, 10, 11}, 10: {9}, 11: {9, 12, 14}, 12: {11, 13, 14},
                    13: {12}, 14: {7, 11, 12}, 15: {6, 16}, 16: {15}}

    reader = TwReader.from_filename(twfile)

    assert reader.num_vertices == 16
    assert reader.num_edges == 36
    assert reader.edges == expected_edges
    assert reader.adjacency_dict == expected_adj



# -*- coding: utf-8 -*-
"""
Testing version.py

Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""
from datetime import date
import re
from pytest import fail
from tdvisu.version import __date__, __version__ as version


def test_semantic_version():
    """Test the version-format with https://semver.org/spec/v2.0.0.html."""
    assert isinstance(version, str)
    regex = (
        r"^(?P<major>0|[1-9]\d*)\."
        r"(?P<minor>0|[1-9]\d*)\."
        r"(?P<patch>0|[1-9]\d*)"
        r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)"
        r"(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))"
        r"?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$")
    result = re.fullmatch(regex, version)
    if result is None:
        fail(
            f"The version {version} does not adhere to semantic versioning!",
            pytrace=False)


def test_date():
    """The date should be 'YYYY-MM-DD' isoformat to be fully portable."""
    assert isinstance(__date__, str)
    # Is it an existing date?
    try:
        date.fromisoformat(__date__)
    except ValueError as err:
        msg = f"""The date {__date__} is not in the format YYYY-MM-DD!
                ValueError: {err}
                Today would be: '{date.today().isoformat()}'"""
        fail(msg, pytrace=False)



# -*- coding: utf-8 -*-
"""
Load and manipulate svg images. Could also be streamed as string.


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

import re
import logging
from pathlib import Path
from typing import Union, Iterable, List, Dict

from benedict import benedict

from tdvisu.utilities import gen_arg

LOGGER = logging.getLogger('svg_join.py')


# indices
WIDTH = 2
HEIGHT = 3


def test_viewbox(viewbox: List[float]):
    """Should be of form [0, 0, +x, +y]"""
    assert len(viewbox) == 4, "viewbox should have exactly 4 values"
    assert viewbox[:2] == [0., 0.], "[min-x,min-y] should be zero."
    assert viewbox[WIDTH] > 0, "should have positive width"
    assert viewbox[HEIGHT] > 0, "should have positive height"


def append_svg(
        first_dict: dict,
        snd_dict: dict,
        centerpad: float = 0.,
        v_bottom: float = None,
        v_top: float = None,
        scale2: float = 1,
        ndigits: int = 3) -> dict:
    """Modifies the first of two xml-svg dictionary containing a viewbox to
    append the second svg to the right of the first image.

    The second svg should only have ONE group 'g'.
    Scaling keeps the top-left corner in place.


    Parameters
    ----------
    first_dict : dict
        Dictionary with key 'svg' including one or more 'g' elements and a
        '@viewBox' attribute.
    snd_dict : dict
        Dictionary with key 'svg' including one 'g' element and a '@viewBox' attribute.
    centerpad : float, optional
        Additional padding in units between the two images. The default is 0.
    v_bottom : float, optional
        Vertical bottomline for the second image relative to the size of the first.
        Can even be negative or greater than one.
    v_top : float, optional
        Vertical top for the second image relative to the size of the first.
        Can even be negative or greater than one.
        If smaller than v_bottom they get swapped.
    scale2 : float, optional
        Optional scaling.
    ndigits : int, optional
        Rounding results to that many decimal places. The default is 3.

    Returns
    -------
    dict
        The extended result in the first_dict.

    """

    first_svg = first_dict['svg']
    second_svg = snd_dict['svg']

    # The value of the viewBox attribute is a list of four numbers:
    #     min-x, min-y, width and height.
    #     The numbers separated by whitespace and/or a comma,
    #     which specify a rectangle in user space which is mapped to the
    #     bounds of the viewport established for the associated SVG element.
    # See also
    # https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/viewBox

    pattern = re.compile(r'\s*,\s*|\s+')
    viewbox1: List[float] = list(
        map(float, re.split(pattern, first_svg['@viewBox'])))
    viewbox2: List[float] = list(
        map(float, re.split(pattern, second_svg['@viewBox'])))

    test_viewbox(viewbox1)  # viewbox1 validation
    test_viewbox(viewbox2)  # viewbox2 validation

    trafo_result = f_transform(
        viewbox1[HEIGHT], viewbox2[HEIGHT], v_bottom, v_top, scale2)
    vertical_snd = trafo_result['vertical_snd']
    combine_height = trafo_result['combine_height']
    scale2 = trafo_result['scale2']

    LOGGER.info(
        "Transformed with vertical_snd=%s combine_height=%s scale2=%s",
        *(vertical_snd, combine_height, scale2))

    viewbox1[HEIGHT] = round(combine_height - 0.5)
    h_displacement = float(viewbox1[WIDTH]) + centerpad
    viewbox1[WIDTH] = round(max(float(viewbox1[WIDTH]),
                                h_displacement + scale2 * viewbox2[WIDTH]) - 0.5)

    # new viewbox
    first_svg['@viewBox'] = ' '.join(map(str, viewbox1))  # new viewbox
    # update width and height
    first_svg['@width'] = f"{viewbox1[WIDTH]}pt"
    first_svg['@height'] = f"{viewbox1[HEIGHT]}pt"
    # move second image group next to first
    v_transform = round(max(0, vertical_snd), ndigits)
    transform = f"translate({h_displacement} {v_transform}) scale({scale2}) "
    # now scales with scale2
    transform += second_svg['g'].get('@transform', '')
    second_svg['g']['@transform'] = transform

    if vertical_snd < 0:
        # move first image, add after other transform
        transform = first_svg['g'].get('@transform', '')
        transform += f" translate(0 {round(-vertical_snd,ndigits)})"
        first_svg['g']['@transform'] = transform

    # add group to list of 'g'
    if isinstance(first_svg['g'], list):
        first_svg['g'].append(second_svg['g'])
    else:
        first_svg['g'] = [first_svg['g'], second_svg['g']]

    return first_dict


TRANSFORMATION_EXAMPLE = """
(0,-0.5)===========> (1,-0.5)
 |
 |              ==========v_top (-0.2)
 |   =========0 |        |
 |   |       |  |        |
 |   | first |  | second |
 |   |       |  |        |
 |   =========1 |        |
 |              ==========v_bottom (1.2)
 V (0,1.3)
"""


def f_transform(h_one_, h_two_,
                v_bottom: Union[float, str, None] = None,
                v_top: Union[float, str, None] = None,
                scale2: float = 1) -> Dict[str, float]:
    """Calculate vertical position and scaling of second image.

    The input for v_bottom, v_top is in units from\n
    0: top of first image\n
    1: bottom of first image\n

    See also the 'transformation_example'!


    Parameters
    ----------
    h_one_ : float-like
        Height of the first image.
    h_two_ : float-like
        Height of the second image.
    v_bottom : float or str, optional
        Expected position of bottom of second image. The default is None.
    v_top : float or str, optional
        Expected position of bottom of second image. The default is None.
    scale2 : float, optional
        Scale the second image. Only used if NOT BOTH v_bottom or v_top are set.

    Returns
    -------
    dict[str->float]
        'vertical_snd','combine_height','scale2'

    """
    v_displacement = 0.
    # cast to float
    h_one = float(h_one_)
    h_two = float(h_two_)
    LOGGER.debug("Calculating with h_one=%f h_two=%f", h_one, h_two)
    # normalize values
    conversion = {'bottom': 1, 'center': 0.5, 'top': 0, 'inf': 0,
                  -float('inf'): 1, float('inf'): 0}
    v_bottom = conversion.get(v_bottom, v_bottom)
    if isinstance(v_bottom, str):
        raise ValueError(f"Encountered {v_bottom=} not in {conversion=}")
    v_top = conversion.get(v_top, v_top)
    if isinstance(v_top, str):
        raise ValueError(f"Encountered {v_top=} not in {conversion=}")

    size2 = h_two * scale2
    # Get v_bottom and v_top calculated (special case None for not set)
    if v_bottom is None and v_top is None:
        v_top = 0  # set to top of first
        v_bottom = v_top + size2 / h_one
    elif v_bottom is None:
        # now top is already set
        v_bottom = v_top + size2 / h_one
    elif v_top is None:
        v_top = v_bottom - size2 / h_one
    else:
        if v_bottom == v_top:
            # moving the centerline according to value and scaling
            LOGGER.info(
                "The values of 'v_top', 'v_bottom' are both interpreted "
                "as %f - interpreting as centerline!", v_top)
            half = size2 / h_one / 2
            v_top = v_top - half
            v_bottom = v_bottom + half
        if v_bottom < v_top:  # swap
            v_top, v_bottom = v_bottom, v_top
        # resulting in scaling
        scale2 = (v_bottom - v_top) * h_one / h_two
        size2 = h_two * scale2

    if size2 < 10:
        LOGGER.warning("Image got scaled down to small size (%s)!", size2)

    v_displacement = v_top * h_one

    # bottom - top
    combine_height = (max(1, v_bottom) - min(0, v_top)) * h_one

    return {'vertical_snd': v_displacement,
            'combine_height': combine_height,
            'scale2': scale2}


def svg_join(
        base_names: list,
        folder: str = '',
        num_images: int = 1,
        outname: str = 'combined',
        suffix: str = '%d.svg',
        preserve_aspectratio: str = 'xMinYMin',
        padding: Union[int, Iterable[int]] = 0,
        scale2: Union[float, Iterable[float]] = 1,
        v_top: Union[None, float, str, Iterable[Union[None, float, str]]] = None,
        v_bottom: Union[None, float, str, Iterable[Union[None, float, str]]] = None):
    """
    Joines different svg-images from tdvisu placed in 'folder' for every timestep
    in the horizontal order specified in 'in_names'.

    Parameters
    ----------
    in_names : list
        Base names of the images to join.
        The method appends the extra 'name'+'%d'+'.svg' to every name with the
        numbering for '%d' starting from 1.
    folder : str, optional
        The working directory. The default is the current directory.
    num_images : int, optional
        Expected maximum for 1<=i<=num_images. The default is only 1.
    outname : str, optional
        The base name for the combined svg. The default is "combined".
        The same timestep and ending '.svg' gets appended to the base name.
    padding : int, optional
        Additional padding in units between every two images. The default is 0.
    preserve_aspectratio : str, optional
        See https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/preserveAspectRatio.
        or https://css-tricks.com/scale-svg/#article-header-id-1 .
        The default is "xMinYMin".
    suffix : str, optional
        Change the prefix for each file. The default is "%d.svg".
    scale2 : float, optional
        Scale the second image. Only used if either v_bottom or v_top is None.
    v_bottom : float or str, optional
        Expected position of bottom of second image. The default is None.
    v_top : float or str, optional
        Expected position of bottom of second image. The default is None.


    Returns
    -------
    None.

    """
    LOGGER.info("Starting joining: %s", base_names)
    # names empty?
    if not base_names:
        LOGGER.warning("svg_join found no images to combine!")
        return
    # only one?
    if len(base_names) == 1:
        LOGGER.warning("svg_join called with one file - nothing to join!")
        return
    # use path library for normalizing the path
    folder = Path(folder if folder else '')

    resultname = str(folder / outname) + suffix
    names = [str(folder / name) + suffix for name in base_names]

    gen_padding = gen_arg(padding)
    gen_scale2 = gen_arg(scale2)
    gen_v_top = gen_arg(v_top)
    gen_v_bottom = gen_arg(v_bottom)

    for step in range(1, num_images + 1):
        # first - needs at least two images
        with open(names[0] % step) as file:
            im_1 = benedict.from_xml(file.read())
        with open(names[1] % step) as file:
            im_2 = benedict.from_xml(file.read())

        result = append_svg(im_1, im_2, centerpad=next(gen_padding),
                            v_bottom=next(gen_v_bottom),
                            v_top=next(gen_v_top), scale2=next(gen_scale2))

        # rest:
        for name in names[2:]:
            with open(name % step) as file:
                image = benedict.from_xml(file.read())
            result = append_svg(result, image, centerpad=next(gen_padding),
                                v_bottom=next(gen_v_bottom),
                                v_top=next(gen_v_top), scale2=next(gen_scale2))

        result['svg']['@preserveAspectRatio'] = preserve_aspectratio
        with open(resultname % step, 'w') as file:
            result.to_xml(output=file, pretty=True)

        if step < 10:
            LOGGER.debug("Wrote combined: %s", resultname % step)
    LOGGER.info("Finished svg_join")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    svg_join(['TDStep', 'graph'],
             'Archive/WheelGraph7',
             outname="default_06sc15_rise",
             v_bottom=[1, .85, .7, .55, .4],
             scale2=1.5,
             num_images=5)



# -*- coding: utf-8 -*-
"""
Helper methods for this module.


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

from collections.abc import Iterable as iter_type
from configparser import ConfigParser, ParsingError, Error as CfgError
import logging
import logging.config
from itertools import chain
from pathlib import Path
from typing import Any, Generator, Iterable, Iterator, List, Tuple, TypeVar, Union
import yaml

LOGGER = logging.getLogger('utilities.py')

CFG_EXT = ('.ini', '.cfg', '.conf', '.config')
LOGLEVEL_EPILOG = """
Logging levels for python 3.8.2:
    CRITICAL: 50
    ERROR:    40
    WARNING:  30
    INFO:     20
    DEBUG:    10
    NOTSET:    0 (will traverse the logging hierarchy until a value is found)
    """
DEFAULT_LOGGING_CFG = {
    'version': 1,
    'formatters': {
        'simple': {
            'format': '%(asctime)s %(levelname)s %(message)s',
            'datefmt': '%H:%M:%S'}},
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'WARNING',
            'formatter': 'simple',
            'stream': 'ext://sys.stdout'}},
    'loggers': {
        'visualization.py': {
            'level': 'NOTSET',
            'handlers': ['console'],
            'propagate': False},
        'svgjoin.py': {
            'level': 'NOTSET',
            'handlers': ['console'],
            'propagate': False},
        'reader.py': {
            'level': 'NOTSET',
            'handlers': ['console'],
            'propagate': False},
        'construct_dpdb_visu.py': {
            'level': 'NOTSET',
            'handlers': ['console'],
            'propagate': False}},
    'root': {
        'level': 'WARNING',
        'handlers': ['console']}}

_T = TypeVar('_T')


def flatten(iterable: Iterable[Iterable[_T]]) -> Iterator[_T]:
    """ Flatten at first level.

    Turn ex=[[1,2],[3,4]] into
    [1, 2, 3, 4]
    and [ex,ex] into
    [[1, 2], [3, 4], [1, 2], [3, 4]]
    """
    return chain.from_iterable(iterable)


def read_yml_or_cfg(file: Union[str, Path], prefer_cfg: bool = False,
                    cfg_ext=CFG_EXT) -> Any:
    """
    Read the file and return its content as a python object.

    Parameters
    ----------
    file : file-like
        The file to read from.
    is_cfg : bool, optional
        Indicate that the file should be in 'config' format.
        Assumed only for ext '.ini', '.cfg', '.conf' and '.config'.
        The default is None.

    Returns
    -------
    Any
        Most likely a dict or a ConfigParser supporting get().
        But maybe just a list or a single object.

    """
    err_str = """utilities.read_yml_or_cfg encountered '{}' while
                reading config from '{}' and prefer_cfg={}"""

    file = Path(file)
    if not file.exists():
        raise FileNotFoundError(file.absolute())
    if not file.is_file():
        raise IsADirectoryError(file.resolve())

    # continue with file
    prefer_cfg = prefer_cfg or file.suffix.lower() in cfg_ext
    if prefer_cfg:
        try:
            config = ConfigParser()
            config.read(file)
            if config.sections():
                return config
            print(err_str.format("empty config", file.resolve(), prefer_cfg))
        except (ParsingError, CfgError) as exc:
            print(err_str.format(exc, file.resolve(), prefer_cfg))

    # try yaml file next
    try:
        return yaml.safe_load(file.open())
    except yaml.error.MarkedYAMLError as exc:
        print(err_str.format(exc, file.resolve(), prefer_cfg))
    if not prefer_cfg:
        try:
            config = ConfigParser()
            config.read(file.open())
            return config
        except (ParsingError, CfgError) as exc:
            print(err_str.format(exc, file.resolve(), prefer_cfg))


def logging_cfg(filename: str, prefer_cfg: bool = False,
                loglevel: Union[None, int, str] = None) -> None:
    """Configure logging for this module"""
    logging.basicConfig()
    read_err = "could not read configuration from '%s'"
    config_err = "could not use logging configuration from '%s'"
    # should be in same directory
    file = Path(__file__).parent / filename

    if loglevel is not None:
        try:
            loglevel = int(float(loglevel))
        except ValueError:
            loglevel = loglevel.upper()

    if prefer_cfg or file.suffix.lower() in CFG_EXT:        # .config
        try:
            logging.config.fileConfig(file, defaults=DEFAULT_LOGGING_CFG)
            if loglevel is not None:
                root = logging.getLogger()
                root.setLevel(loglevel)
                for handler in root.handlers:
                    handler.setLevel(loglevel)
            return
        except OSError:
            LOGGER.error(read_err, file.resolve(), exc_info=True)
        except ValueError:
            LOGGER.error(config_err, file.resolve(), exc_info=True)
    try:                                                    # dict
        file_content = read_yml_or_cfg(file, prefer_cfg=prefer_cfg)
        logging.config.dictConfig(file_content)
        if loglevel is not None:
            root = logging.getLogger()
            root.setLevel(loglevel)
            for handler in root.handlers:
                handler.setLevel(loglevel)
        return
    except OSError:
        LOGGER.error(read_err, file.resolve(), exc_info=True)
    except ValueError:
        LOGGER.error(config_err, file.resolve(), exc_info=True)


def convert_to_adj(
        edgelist: Iterable[Tuple[int, int]], directed: bool = False) -> dict:
    """
    Helper function to convert the edgelist into the adj-format from NetworkX.

    Parameters
    ----------
    edgelist : array-like of pairs of vertices.
        Simple edgelist. Example:
            [(2, 1), (3, 2), (4, 2), (5, 4)]
    directed : bool, optional
        Whether to add the backward edges too. The default is False.

    Returns
    -------
    adj : dict of edges with empty attributes
        See Docs » Module code » networkx.classes.graph.adj(self)
        for detailed structure.
        Basically: dict of {source1:{target1:{'attr1':value,},},...}
        https://networkx.github.io/documentation/networkx-2.1/_modules/networkx/classes/graph.html
    """
    adj = dict()
    for (source, target) in edgelist:
        if source not in adj:
            adj[source] = {}
        adj[source][target] = {}
        if not directed:
            # add reversed edge
            if target not in adj:
                adj[target] = {}
            adj[target][source] = {}
    return adj


def add_edge_to(
        edges: set,
        adjacency_dict: dict,
        vertex1: Any,
        vertex2: Any) -> None:
    """
    Adding (undirected) edge from 'vertex1' to 'vertex2'
    to the edges and adjacency-list.

    Parameters
    ----------
    edges : set-like
        Set of tuples of vertices.
    adjacency_dict : dict-like
        Saves adjecent vertices for each vertex.
    vertex1 : any
        First vertex of the new edge.
    vertex2 : any
        Second vertex of the new edge.

    Returns
    -------
    None

    """
    if vertex1 in adjacency_dict:
        adjacency_dict[vertex1].add(vertex2)
    else:
        adjacency_dict[vertex1] = {vertex2}
    if vertex2 in adjacency_dict:
        adjacency_dict[vertex2].add(vertex1)
    else:
        adjacency_dict[vertex2] = {vertex1}
    edges.add((vertex1, vertex2))


def gen_arg(arg_or_iter: Any) -> Generator:
    """
    Infinite generator for the next argument of `arg_or_iter`.
    If the argument is exhausted, always return the last element.

    Parameters
    ----------
    arg_or_iter : object
        Object to iterate over. Considers three cases:
            string: yields the string as one element indefinitely
            iterable: yields all elements from it, and only the last one after.
            not iterable: yield the object indefinitely
    """
    if isinstance(arg_or_iter, str):
        while True:
            yield arg_or_iter
    elif not isinstance(arg_or_iter, iter_type):
        while True:
            yield arg_or_iter
    else:
        item = None
        for item in arg_or_iter:
            yield item
    while True:
        yield item


def base_style(
        graph,
        node: str,
        color: str = 'white',
        penwidth: float = 1.0) -> None:
    """Style the node with default fillcolor and penwidth."""
    graph.node(node, fillcolor=color, penwidth=str(penwidth))


def emphasise_node(graph, node: str, color: str = 'yellow',
                   penwidth: float = 2.5) -> None:
    """Emphasise node with a different fillcolor (default:'yellow')
    and penwidth (default:2.5).
    """
    if color:
        graph.node(node, fillcolor=color)
    if penwidth:
        graph.node(node, penwidth=str(penwidth))


def style_hide_node(graph, node: str) -> None:
    """Make the node invisible during drawing."""
    graph.node(node, style='invis')


def style_hide_edge(graph, source: str, target: str) -> None:
    """Make the edge source->target invisible during drawing."""
    graph.edge(source, target, style='invis')


def bag_node(
        head,
        tail,
        anchor: str = 'anchor',
        headcolor: str = 'white',
        tableborder: int = 0,
        cellborder: int = 0,
        cellspacing: int = 0) -> str:
    """HTML format with 'head' as the first label, then appending
    further labels.

    After the 'head' there is an (empty) anchor for edges with a name tag. e.g.
    <TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0">
    <TR><TD BGCOLOR="white">bag 3</TD></TR><TR><TD PORT="anchor"></TD></TR>
    <TR><TD>[1, 2, 5]</TD></TR><TR><TD>03/31/20 09:29:51</TD></TR>
    <TR><TD>dtime=0.0051s</TD></TR></TABLE>
    """
    result = f"""<<TABLE BORDER=\"{tableborder}\" CELLBORDER=\"{cellborder}\"
              CELLSPACING=\"{cellspacing}\">
              <TR><TD BGCOLOR=\"{headcolor}\">{head}</TD></TR>
              <TR><TD PORT=\"{anchor}\"></TD></TR>"""

    if isinstance(tail, str):
        result += f"<TR><TD>{tail}</TD></TR>"
    else:
        for label in tail:
            result += f"<TR><TD>{label}</TD></TR>"

    result += "</TABLE>>"
    return result


def solution_node(
        solution_table: Iterable[List[str]],
        toplabel: str = '',
        bottomlabel: str = '',
        transpose: bool = False,
        linesmax: int = 1000,
        columnsmax: int = 50) -> str:
    """Fill the node from the 2D 'solution_table' (columnbased!).
    Optionally add a line above and/or below the table.

    solution_table : 2D-arraylike, entries get converted to str

    toplabel : string, placed above the table

    bottomlabel : string, placed below the table

    transpose : bool, whether to transpose the solution_table before
    processing

    linesmax : int, if positive it indicates the
            maximum number of lines in the table to display.

    columnsmax : int, if positive it indicates the
            maximum number of columns to display + the last.

    Example structure for four columns:
    |----------|
    | toplabel |
    ------------
    |v1|v2|v3|v4|
    |0 |1 |0 |1 |
    |1 |1 |0 |0 |
    ...
    ------------
    | botlabel |
    |----------|
    """
    result = ''
    if toplabel:
        result += toplabel + '|'

    if len(solution_table) == 0:
        result += 'empty'
    else:
        if transpose:
            solution_table = list(zip(*solution_table))

        # limit lines backwards from length of column
        vslice = (min(-1, linesmax - len(solution_table[0]))
                  if linesmax > 0 else -1)
        # limit columns forwards minus one
        hslice = (min(len(solution_table), columnsmax)
                  if columnsmax > 0 else len(solution_table)) - 1

        result += '{'                                       # insert table
        for column in solution_table[:hslice]:
            result += '{'                                   # start column
            for row in column[:vslice]:
                result += str(row) + '|'
            if vslice < -1:     # add one indicator of shortening
                result += '...' + '|'
            for row in column[-1:]:
                result += str(row)
            result += '}|'      # sep. between columns
        # adding one column-skipping indicator
        if hslice < len(solution_table) - 1:
            result += '{'                                   # start column
            for row in column[:vslice]:
                result += '...' + '|'
            if vslice < -1:     # add one indicator of shortening
                result += '...' + '|'
            for row in column[-1:]:
                result += '...'
            result += '}|'      # sep. between columns
        # last column (usually a summary of the previous cols)
        for column in solution_table[-1:]:
            result += '{'                                   # start column
            for row in column[:vslice]:
                result += str(row) + '|'
            if vslice < -1:     # add one indicator of shortening
                result += '...' + '|'
            for row in column[-1:]:
                result += str(row)
            result += '}'      # sep. between columns
        result += '}'                                       # close table

    if bottomlabel:
        result += '|' + bottomlabel

    return '{' + result + '}'



# -*- coding: utf-8 -*-
"""
Bidirectional Dijkstra's Algorithm for finding the shortest path between nodes in a graph.
For more information on the bidirectional agorithm see for example
cs.princeton.edu/courses/archive/spr06/cos423/Handouts/EPP%20shortest%20path%20algorithms.pdf


Copyright (C) 2020 Martin Röbke <martin.roebke@tu-dresden.de>
Modified sourcecode from NetworkX.

Copyright (C) 2004-2020, NetworkX Developers
Aric Hagberg <hagberg@lanl.gov>
Dan Schult <dschult@colgate.edu>
Pieter Swart <swart@lanl.gov>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

  * Neither the name of the NetworkX Developers nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


from heapq import heappush, heappop
from itertools import count


def bidirectional_dijkstra(edges, source, target, weight='weight'):
    r"""Dijkstra's algorithm for shortest paths using bidirectional search.

    Parameters
    ----------
    edges : dict of edges
        See Docs » Module code » networkx.classes.graph.adj(self)
        for detailed structure.
        Basically: dict of {source1:{target1:{'attr1':value,},},...}
        networkx.github.io/documentation/networkx-2.1/_modules/networkx/classes/graph.html

    source : node
       Starting node.

    target : node
       Ending node.

    weight : string or function,
       If this is a string, then edge weights will be accessed via the
       edge attribute with this key (that is, the weight of the edge
       joining `u` to `v` will be ``G.edges[u, v][weight]``). If no
       such edge attribute exists, the weight of the edge is assumed to
       be one.

       If this is a function, the weight of an edge is the value
       returned by the function. The function must accept exactly three
       positional arguments: the two endpoints of an edge and the
       dictionary of edge attributes for that edge. The function must
       return a number.

    Returns
    -------
    length, path : number and list
       length is the distance from source to target.
       path is a list of nodes on a path from source to target.

    Raises
    ------
    ValueError
        If either `source` or `target` is not in edges.

    DijkstraNoPath
        If no path exists between source and target.

    Examples
    --------


    Notes
    -----
    Edge weight attributes must be numerical.
    Distances are calculated as sums of weighted edges traversed.

    In practice  bidirectional Dijkstra is much more than twice as fast as
    ordinary Dijkstra.

    Ordinary Dijkstra expands nodes in a sphere-like manner from the
    source. The radius of this sphere will eventually be the length
    of the shortest path. Bidirectional Dijkstra will expand nodes
    from both the source and the target, making two spheres of half
    this radius. Volume of the first sphere is `\pi*r*r` while the
    others are `2*\pi*r/2*r/2`, making up half the volume.

    This algorithm is not guaranteed to work if edge weights
    are negative or are floating point numbers
    (overflows and roundoff errors can cause problems).

    """
    if source not in edges or target not in edges:
        msg = f"Either source {source} or target {target} is not in edges"
        raise ValueError(msg)

    if source == target:
        return (0, [source])

    weight = _weight_function(weight)
    push = heappush
    pop = heappop
    # Init:  [Forward, Backward]
    dists = [{}, {}]   # dictionary of final distances
    paths = [{source: [source]}, {target: [target]}]  # dictionary of paths
    fringe = [[], []]  # heap of (distance, node) for choosing node to expand
    seen = [{source: 0}, {target: 0}]  # dict of distances to seen nodes

    # initialize fringe heap
    push(fringe[0], (0, next(count()), source))
    push(fringe[1], (0, next(count()), target))
    # neighs for extracting correct neighbor information
    neighs = [edges, edges]
    # variables to hold shortest discovered path
    finaldist = float("inf")
    finalpath = []
    direction = 1
    while fringe[0] and fringe[1]:
        # choose direction
        # direction == 0 is forward direction and direction == 1 is back
        direction = 1 - direction
        # extract closest to expand
        (dist, _, v) = pop(fringe[direction])
        if v in dists[direction]:
            # Shortest path to v has already been found
            continue
        # update distance
        dists[direction][v] = dist  # equal to seen[direction][v]
        if v in dists[1 - direction]:
            # if we have scanned v in both directions we are done
            # we have now discovered the shortest path
            return (finaldist, finalpath)

        for w, d in neighs[direction][v].items():
            if direction == 0:  # forward
                vw_length = dists[direction][v] + weight(v, w, d)
            else:  # back, must remember to change v,w->w,v
                vw_length = dists[direction][v] + weight(w, v, d)
            if w in dists[direction]:
                if vw_length < dists[direction][w]:
                    raise ValueError(
                        "Contradictory paths found: negative weights?")
            elif w not in seen[direction] or vw_length < seen[direction][w]:
                # relaxing
                seen[direction][w] = vw_length
                push(fringe[direction], (vw_length, next(count()), w))
                paths[direction][w] = paths[direction][v] + [w]
                if w in seen[0] and w in seen[1]:
                    # see if this path is better than than the already
                    # discovered shortest path
                    totaldist = seen[0][w] + seen[1][w]
                    if finalpath == [] or finaldist > totaldist:
                        finaldist = totaldist
                        revpath = paths[1][w][:]
                        revpath.reverse()
                        finalpath = paths[0][w] + revpath[1:]
    raise DijkstraNoPath(f"No path between {source} and {target}.")


class DijkstraNoPath(RuntimeError):
    """Raised when there was no path found during Dijkstra algorithm"""


def _weight_function(weight, multigraph: bool = False):
    """Returns a function that returns the weight of an edge.

    The returned function is specifically suitable for input to
    functions :func:`_dijkstra` and :func:`_bellman_ford_relaxation`.

    Parameters
    ----------
    weight : string or function
        If it is callable, `weight` itself is returned. If it is a string,
        it is assumed to be the name of the edge attribute that represents
        the weight of an edge. In that case, a function is returned that
        gets the edge weight according to the specified edge attribute.

    multigraph : bool
        Whether the edges represent a multigraph.
        The default is False.

    Returns
    -------
    function
        This function returns a callable that accepts exactly three inputs:
        a node, an node adjacent to the first one, and the edge attribute
        dictionary for the edge joining those nodes. That function returns
        a number representing the weight of an edge.

    If multigraph is true, and `weight` is not callable, the
    minimum edge weight over all parallel edges is returned. If any edge
    does not have an attribute with key `weight`, it is assumed to
    have weight one.

    """
    if callable(weight):
        return weight
    # If the weight keyword argument is not callable, we assume it is a
    # string representing the edge attribute containing the weight of
    # the edge.
    if multigraph:
        return lambda u, v, d: min(edge.get(weight, 1) for edge in d.values())
    return lambda u, v, data: data.get(weight, 1)


if __name__ == "__main__":
    # Show one example and print to console
    EDGES = {2: {1: {}, 3: {}, 4: {}},
             1: {2: {}},
             3: {2: {}},
             4: {2: {}, 5: {}},
             5: {4: {}}}
    RESULT = bidirectional_dijkstra(EDGES, 3, 5)
    print(RESULT)



# -*- coding: utf-8 -*-
"""
Visualization Data


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

from typing import Union, Iterable, Optional
from dataclasses import dataclass


@dataclass
class SvgJoinData:
    """Class for holding different parameters to join the results."""
    base_names: Union[str, Iterable[str]]
    folder: Optional[str] = None
    outname: str = 'combined'
    suffix: str = '%d.svg'
    preserve_aspectratio: str = 'xMinYMin'
    num_images: int = 1
    padding: Union[int, Iterable[int]] = 0
    scale2: Union[float, Iterable[float]] = 1.0
    v_top: Union[None, float, str, Iterable[Union[None, float, str]]] = None
    v_bottom: Union[None, float, str, Iterable[Union[None, float, str]]] = None


@dataclass
class IncidenceGraphData:
    """Class holding different parameters for the incidence graph."""
    edges: list
    subgraph_name_one: str = 'clauses'
    subgraph_name_two: str = 'variables'
    var_name_one: str = ''
    var_name_two: str = ''
    infer_primal: bool = False
    infer_dual: bool = False
    primal_file: str = 'PrimalGraphStep'
    inc_file: str = 'IncidenceGraphStep'
    dual_file: str = 'DualGraphStep'
    fontsize: int = 16
    penwidth: float = 2.2
    second_shape: str = 'diamond'
    column_distance: float = 0.5


@dataclass
class GeneralGraphData:
    """Class holding different parameters for the general graph."""
    edges: list
    extra_nodes: Optional[list] = None
    graph_name: str = 'graph'
    file_basename: str = 'graph'
    var_name: str = ''
    do_sort_nodes: bool = False
    do_adj_nodes: bool = False
    fontsize: int = 20
    first_color: str = 'yellow'
    first_style: str = 'filled'
    second_color: str = 'green'
    second_style: str = 'dotted,filled'
    third_color: str = 'red'

    def __post_init__(self):
        if self.extra_nodes is None:
            self.extra_nodes = list()


@dataclass
class VisualizationData:
    """Class holding different parameters for Visualization."""
    incidence_graph: Optional[IncidenceGraphData] = None
    general_graph: Optional[GeneralGraphData] = None
    svg_join: Optional[SvgJoinData] = None
    td_file: str = 'TDStep'
    colors: Optional[list] = None
    orientation: str = 'BT'
    linesmax: int = 100
    columnsmax: int = 20
    bagcolor: str = 'white'
    fontsize: int = 20
    penwidth: float = 2.2
    fontcolor: str = 'black'
    emphasis: Optional[dict] = None

    def __post_init__(self):
        if self.colors is None:
            self.colors = [
                '#0073a1',
                '#b14923',
                '#244320',
                '#b1740f',
                '#a682ff',
                '#004066',
                '#0d1321',
                '#da1167',
                '#604909',
                '#0073a1',
                '#b14923',
                '#244320',
                '#b1740f',
                '#a682ff']
        if self.emphasis is None:
            self.emphasis = dict()
        # merge input over defaults:
        self.emphasis = {**{"firstcolor": 'yellow',
                            "secondcolor": 'green',
                            "firststyle": 'filled',
                            "secondstyle": 'dotted,filled'
                            },
                         **self.emphasis}


if __name__ == "__main__":
    # Just Testing:
    incid = IncidenceGraphData([])
    gen = GeneralGraphData([])
    data = VisualizationData(incidence_graph=incid, general_graph=gen)
    print(data)



# -*- coding: utf-8 -*-
"""
Testing svgjoin.py


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

from os import makedirs
from os.path import dirname, join
from random import randint
from typing import Generator
from collections.abc import Iterable as iter_type
from pytest import param, mark
from benedict import benedict

from tdvisu.svgjoin import f_transform, append_svg, gen_arg


WRITE = False  # ??? Write Testimages instead of just reading them ???

IMAGE_FOLDER = 'expected_images'
DIR = join(dirname(__file__), IMAGE_FOLDER)
FILE1 = join(dirname(__file__), 'IncidenceGraphStep11.svg')
FILE2 = join(dirname(__file__), 'PrimalGraphStep11.svg')

MIN, MAX = 5, 1_000_000
# Should be in between sizes considered for image-dimensions!
BASE = randint(20, 3000)
last_random = None


def randomized(lower: int = MIN, upper: int = MAX):
    """Return randint(lower, upper) with the default values
    considered for image dimensions.
    """
    return randint(lower, upper)


def rand_larger(number):
    """Return a random larger image-size than 'number' and save in last_random."""
    global last_random
    last_random = randomized(lower=number + 1)
    return last_random


def rand_smaller(number):
    """Return a random smaller image-size than 'number' and save in last_random."""
    global last_random
    last_random = randomized(upper=number - 1)
    return last_random


class TestNewHeight:
    """Test the transform method in svgjoin"""

    parameters_default = [
        param({'h_one_': BASE, 'h_two_': BASE},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id='Only heights (same)'),
        param({'h_one_': BASE, 'h_two_': 2 * BASE},
              {'vertical_snd': 0.0, 'combine_height': 2 * BASE,
               'scale2': 1}, id='Only heights (larger)'),
        param({'h_one_': BASE, 'h_two_': 0.5 * BASE},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id='Only heights (smaller)'),
        param({'h_one_': BASE, 'h_two_': BASE, 'v_bottom': None},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id='Default v_bottom'),
        param({'h_one_': BASE, 'h_two_': BASE, 'v_bottom': None, 'v_top': None},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id='Default v_bottom&v_top'),
        param({'h_one_': BASE, 'h_two_': BASE, 'scale2': 1},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id='Default scale2')]

    parameters_moving = [
        param({'h_one_': BASE, 'h_two_': BASE, 'v_bottom': 1, 'v_top': 0},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id='static'),
        param({'h_one_': BASE, 'h_two_': BASE, 'v_bottom': 0, 'v_top': 1},
              {'vertical_snd': 0.0, 'combine_height': BASE,
               'scale2': 1}, id="switched bottom and top -> "
              "should switch automatically!"),
        param({'h_one_': BASE, 'h_two_': rand_smaller(BASE), 'v_bottom': 1, 'v_top': 0},
              {'vertical_snd': 0.0,
               'combine_height': BASE, 'scale2': BASE / last_random}, id='scale up to BASE'),
        param({'h_one_': BASE, 'h_two_': rand_larger(BASE), 'v_bottom': 1, 'v_top': 0},
              {'vertical_snd': 0.0,
               'combine_height': BASE, 'scale2': BASE / last_random}, id='scale down to BASE'),
    ]

    @mark.parametrize("arguments,expected", parameters_default)
    def test_parameters_default(self, arguments, expected):
        """Test that the default parameters from f_transform work as expected."""
        result = f_transform(**arguments)
        assert result == expected

    @mark.parametrize("arguments,expected", parameters_moving)
    def test_parameters_moving(self, arguments, expected):
        """Test that different parameters for f_transform work as expected."""
        result = f_transform(**arguments)
        assert result == expected


@mark.parametrize(
    "arg",
    [param(None,
           id="Should accept None as argument"),
     param(randint(1, 100),
           id="One integer in generator"),
     param("Hello Test",
           id="Should yield the string itself"),
     param([None, randint(1, 1000), None],
           id="Should accept None as element in list"),
     param([[1, 2], "String", None, [[[1]]]],
           id="Should only return elements from the first level")
     ]
)
def test_gen_arg(arg):
    """Test the generator in svgjoin"""
    gen = gen_arg(arg)
    assert isinstance(gen, Generator)
    size = randint(10, 40)
    if isinstance(arg, str) or not isinstance(arg, iter_type):
        assert [next(gen) for _ in range(size)] == [arg for _ in range(size)]
    else:
        assert ([next(gen) for _ in range(size)] ==
                arg + [arg[-1] for _ in range(size - len(arg))])


@mark.parametrize(
    "otherargs, filename, reverse",
    [param(dict(), 'result_simple_join', False,
           id="Combine two example svg images to a new one"),
     param(dict(), 'result_simple_join', True,
           id="Test the reverse order"),
     param(dict(centerpad=100), 'result_simple_join_padding', True,
           id="Add horizontal padding"),
     param(dict(v_bottom='center', v_top='center'), 'result_centered_join', False,
           id="Center the image"),
     param(dict(v_bottom='center', v_top='center'), 'result_centered_join', True,
           id="Center the image and reversed")
     ]
)
@mark.parametrize(
    "scale2",
    [1, 2, 0.5, 0.2]
)
def test_append_svg_scaled(otherargs, filename, reverse, scale2, write=WRITE):
    """Combine two example svg images - different scalings for second image."""
    filename += f"_scale{int(scale2*10)}to10_{reverse=}.svg"
    with open(FILE1) as file1:
        im_1 = benedict.from_xml(file1.read())
        with open(FILE2) as file2:
            im_2 = benedict.from_xml(file2.read())
            order = (im_2, im_1) if reverse else (im_1, im_2)
            result = append_svg(*order, scale2=scale2, **otherargs)

            result['svg']['@preserveAspectRatio'] = "xMinYMin"
            if write:
                makedirs(DIR, exist_ok=True)
                with open(join(DIR, filename), 'w') as outfile:
                    result.to_xml(output=outfile, pretty=True)
            with open(join(DIR, filename), 'r') as expected:
                assert result == benedict.from_xml(expected.read())


@mark.parametrize(
    "otherargs, filename, reverse",
    [
        param(dict(v_bottom=0, v_top=1), 'result_same_size', False,
              id="Scale second to same size"),
        param(dict(v_bottom=0, v_top=1), 'result_same_size', True,
              id="Scale first to same size"),
        param(dict(v_bottom='bottom', v_top='center'), 'result_lower_half', False,
              id="Move to lower half"),
        param(dict(v_bottom='center', v_top='top'), 'result_upper_half', False,
              id="Move to upper half"),
        param(dict(v_bottom='center', v_top='top'), 'result_upper_half', True,
              id="Move to upper half and reversed"),
        param(dict(v_bottom='top', v_top=-.5), 'result_top_up', False,
              id="Move to top_-0.5"),
        param(dict(v_bottom=0.2, v_top=-.3), 'result_0p2_up', False,
              id="Move to 0.2_-0.3"),
        param(dict(v_bottom='bottom', v_top=1.5), 'result_1p5_down', False,
              id="Move below bottom_1.5"),
    ]
)
@mark.parametrize(
    "padding",
    [0, 50, 200]
)
def test_append_svg_pinned(otherargs, filename, reverse, padding, write=WRITE):
    """Combine two example svg images - scaling with v_bottom and v_top."""
    filename += f"_padding{padding}_{reverse=}.svg"
    with open(FILE1) as file1:
        im_1 = benedict.from_xml(file1.read())
        with open(FILE2) as file2:
            im_2 = benedict.from_xml(file2.read())
            order = (im_2, im_1) if reverse else (im_1, im_2)
            result = append_svg(*order, centerpad=padding, **otherargs)

            result['svg']['@preserveAspectRatio'] = "xMinYMin"
            if write:
                makedirs(DIR, exist_ok=True)
                with open(join(DIR, filename), 'w') as outfile:
                    result.to_xml(output=outfile, pretty=True)
            with open(join(DIR, filename), 'r') as expected:
                assert result == benedict.from_xml(expected.read())



# -*- coding: utf-8 -*-
"""
Testing creation of bag strings in visualization.py

Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

from pytest import param, mark
from tdvisu.utilities import solution_node


SOLUTIONTABLE1 = [["id", "0"], ["v1", "1"],
                  ["v2", "2"], ["v3", "4"], ["nSol", "0"]]

SOLUTIONTABLEINT = [["id", 0], ["v1", 1], ["v2", 2], ["v3", 4], ["nSol", 0]]

SOLUTIONTABLEFLOAT = [["id", 0.1], ["v1", 1.], ["v2", 2.2222], ["v3", 4.4],
                      ["nSol", 0.1]]

SOLUTIONHEADER = [["id"], ["v1"], ["v2"], ["v3"], ["nSol"]]


parameters_sol = [
    param({'solution_table': tuple()},
          "{empty}", id='with empty args'),
    param({'solution_table': tuple(), 'toplabel': 'top'},
          "{top|empty}", id='with one toplabel'),
    param({'solution_table': tuple(), 'bottomlabel': 'bottom'},
          "{empty|bottom}", id='with one bottomlabel'),
    param({'solution_table': SOLUTIONTABLE1, 'toplabel': 'top',
           'bottomlabel': 'bottom', 'transpose': True},
          "{top|{{id|v1|v2|v3|nSol}|{0|1|2|4|0}}|bottom}",
          id='with both labels and transposed'),
    param({'solution_table': SOLUTIONTABLE1},
          "{{{id|0}|{v1|1}|{v2|2}|{v3|4}|{nSol|0}}}",
          id='not transposed no label'),
    param({'solution_table': SOLUTIONTABLE1, 'toplabel': 'top',
           'bottomlabel': 'bottom'},
          "{top|{{id|0}|{v1|1}|{v2|2}|{v3|4}|{nSol|0}}|bottom}",
          id='not transposed plus labels'),
    param({'solution_table': SOLUTIONHEADER},
          "{{{id}|{v1}|{v2}|{v3}|{nSol}}}",
          id='only header'),
    param({'solution_table': SOLUTIONHEADER, 'toplabel': 'top',
           'bottomlabel': 'bottom'},
          "{top|{{id}|{v1}|{v2}|{v3}|{nSol}}|bottom}",
          id='only header plus labels'),
    param({'solution_table': SOLUTIONTABLEINT},
          "{{{id|0}|{v1|1}|{v2|2}|{v3|4}|{nSol|0}}}",
          id='conversion of ints in table'),
    param({'solution_table': SOLUTIONTABLEFLOAT},
          "{{{id|0.1}|{v1|1.0}|{v2|2.2222}|{v3|4.4}|{nSol|0.1}}}",
          id='conversion of floats'), ]


@mark.parametrize("arguments,expected", parameters_sol)
def test_solutionnode(arguments, expected):
    """Testing different solution_node capabilities."""

    result = solution_node(**arguments)
    assert result == expected



# -*- coding: utf-8 -*-
"""
Testing visualization.py


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

import argparse
from pathlib import Path

from tdvisu.visualization import main


def test_sat_and_join(tmpdir):
    """Complete visualization run with svgjoin."""
    parser = argparse.ArgumentParser()
    parser.add_argument('infile',
                        type=argparse.FileType('r', encoding='UTF-8'),
                        help="Input file for the visualization "
                        "must conform with the 'JsonAPI.md'")
    parser.add_argument('outfolder',
                        help="Folder to output the visualization results")
    parser.add_argument('--loglevel', help="set the minimal loglevel for root")

    # get cmd-arguments
    infile = Path(__file__).parent / 'dbjson4andjoin.json'
    outfolder = Path(tmpdir) / 'temp-test_sat_and_join'
    _args = parser.parse_args([str(infile), str(outfolder)])
    # call main()
    main(_args)
    files = [file for file in outfolder.iterdir() if file.is_file()]
    assert len(files
               ) == 42, "total files"
    assert len([file for file in files if file.suffix == '.svg']
               ) == 24, "svg files"
    assert len([file for file in files if file.suffix == '']
               ) == 18, "dot source files"

    testobjects = [file + str(i) for file in
                   ["IncidenceGraphStep", "PrimalGraphStep", "TDStep"]
                   for i in range(1, 7)]

    expected = [Path(__file__).parent / 'expected_images' / file
                for file in testobjects]

    for file in testobjects:
        with open(Path(__file__).parent / 'expected_images' / file) as expected:
            with open(outfolder / file) as result:
                assert result.read() == expected.read(
                ), f"{file} should be the same"



# -*- coding: utf-8 -*-
"""
Testing utilities.py

Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""
from pytest import param, mark, raises
from tdvisu.utilities import flatten, convert_to_adj, add_edge_to


@mark.parametrize(
    "arg",
    [None, 10, 1.4, [1, 2, 3], [None]]
)
def test_cant_flatten(arg):
    """Fail the flatten method with TypeError"""
    with raises(TypeError):
        list(flatten(arg))


@mark.parametrize(
    "arg, expected",
    [((), []),
     ([], []),
     (([None],), [None]),
     ([(1, 2), (3, 4)], [1, 2, 3, 4]),
     ([(1, 2)], [1, 2]),
     ([""], []),
     (["String"], ['S', 't', 'r', 'i', 'n', 'g']),
     (["Str", "Zw", "Dr"], ['S', 't', 'r', 'Z', 'w', 'D', 'r'])
     ]
)
def test_flatten(arg, expected):
    """Test the flatten method."""
    assert list(flatten(arg)) == expected


def test_convert_to_adj():
    """Test the convert_to_adj method"""
    assert convert_to_adj([(2, 1), (3, 2), (4, 2), (5, 4)]) == {
        2: {1: {}, 3: {}, 4: {}}, 1: {2: {}}, 3: {2: {}}, 4: {2: {}, 5: {}}, 5: {4: {}}}


@mark.parametrize(
    "edges, adj, vertex1, vertex2, new_adj",
    [param(set(), {}, 1, 2, {1: {2}, 2: {1}},
           id="empty sets 1"),
     param(set(), {}, "B", "A", {"B": {"A"}, "A": {"B"}},
           id="empty sets 2"),
     param(set([1, 2, 3]), {}, 1, 2, {1: {2}, 2: {1}},
           id="doesnt care about prev elements 1"),
     param(set([None]), {}, "B", "A", {"B": {"A"}, "A": {"B"}},
           id="doesnt care about prev elements 2"),
     param({(1, 2), (2, 3)}, {1: {2}, 2: {1, 3}, 3: {2}}, 3, 1,

           {1: {2, 3}, 2: {1, 3}, 3: {1, 2}},
           id="Closing the triangle"),
     param({('a', 1), (1, 2), (4.3, 2)}, {1: {2, 'a'}, 2: {1, 4.3}, 4.3: {2}, 'a': {1}},
           2, "a",

           {1: {2, 'a'}, 2: {1, 4.3, 'a'}, 4.3: {2}, 'a': {1, 2}},
           id="Even with different elements everything works"),
     ]
)
def test_add_edge_to(edges, adj, vertex1, vertex2, new_adj):
    """Test the add_edge_to method. Sensitive to order in vertices in edges!"""
    expect_edges = set(edges)
    expect_edges.add((vertex1, vertex2))
    add_edge_to(edges, adj, vertex1, vertex2)
    assert edges == expect_edges
    assert adj == new_adj
    # adding second time switched does not change the adjacency
    add_edge_to(edges, adj, vertex2, vertex1)
    assert adj == new_adj



# -*- coding: utf-8 -*-
"""
Construct visualization-JSON from dpdb-database results.

See https://www.postgresqltutorial.com/postgresql-python/connect/
and reference
https://github.com/VaeterchenFrost/dp_on_dbs.git

Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""
import abc
import argparse
import json
import logging
from pathlib import Path
from time import sleep

import psycopg2 as pg

from tdvisu.dijkstra import bidirectional_dijkstra as find_path
from tdvisu.utilities import convert_to_adj
from tdvisu.reader import TwReader
from tdvisu.visualization import flatten
from tdvisu.version import __date__, __version__ as version
from tdvisu.utilities import read_yml_or_cfg, logging_cfg, LOGLEVEL_EPILOG

LOGGER = logging.getLogger('construct_dpdb_visu.py')

DEFAULT_DBCONFIG = {
    "host": "localhost",
    "port": 5432,
    "database": "logicsem",
    "user": "postgres",
    "application_name": "dpdb-admin"
}

PSYCOPG2_8_5_TASTATUS = {
    pg.extensions.TRANSACTION_STATUS_IDLE:
        ('TRANSACTION_STATUS_IDLE ',
         '(The session is idle and there is no current transaction.)'),

        pg.extensions.TRANSACTION_STATUS_ACTIVE:
        ('TRANSACTION_STATUS_ACTIVE ',
         '(A command is currently in progress.)'),

        pg.extensions.TRANSACTION_STATUS_INTRANS:
        ('TRANSACTION_STATUS_INTRANS ',
         '(The session is idle in a valid transaction block.)'),

        pg.extensions.TRANSACTION_STATUS_INERROR:
        ('TRANSACTION_STATUS_INERROR ',
         '(The session is idle in a failed transaction block.)'),

        pg.extensions.TRANSACTION_STATUS_UNKNOWN:
        ('TRANSACTION_STATUS_UNKNOWN ',
         '(Reported if the connection with the server is bad.)')
}


def good_db_status() -> tuple:
    """Any good db status to proceed."""
    return (pg.extensions.TRANSACTION_STATUS_IDLE,
            pg.extensions.TRANSACTION_STATUS_INTRANS)


def read_cfg(cfg_file, section: str, prefer_cfg: bool = False) -> dict:
    """Read the config file and return the result of one section."""
    try:
        file_content = read_yml_or_cfg(cfg_file, prefer_cfg=prefer_cfg)
        content = dict(file_content[section])
        LOGGER.debug(
            "Found keys %s in %s[%s]", content.keys(), cfg_file, section)
    except (OSError, AttributeError, TypeError) as err:
        LOGGER.warning("Encountered %s while reading config '%s' section '%s'",
                       err, cfg_file, section, exc_info=True)
        content = {}
    return content


def db_config(filename: str = 'database.ini',
              section: str = 'postgresql') -> dict:
    """Return the database config as JSON"""
    file = Path(__file__).parent / filename
    LOGGER.info("Read db_config['%s'] from '%s'", section, file.resolve())
    cfg = read_cfg(file, section)
    return {**DEFAULT_DBCONFIG, **cfg}


class IDpdbVisuConstruct(metaclass=abc.ABCMeta):
    """Interface for parsing database results from dynamic programming
    into the JSON used for visualizing the solution steps
    on the tree decomposition.
    See details for i-face impl in https://realpython.com/python-interface/
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'construct') and
                callable(subclass.construct) and
                hasattr(subclass, 'read_labeldict') and
                callable(subclass.read_labeldict) and
                hasattr(subclass, 'read_timeline') and
                callable(subclass.read_timeline) and
                hasattr(subclass, 'read_edgearray') and
                callable(subclass.read_edgearray) or
                NotImplemented)

    @abc.abstractmethod
    def construct(self) -> dict:
        """Return the constructed Json."""
        raise NotImplementedError

    @abc.abstractmethod
    def read_edgearray(self) -> list:
        """Return the edges between the bags."""
        raise NotImplementedError

    @abc.abstractmethod
    def read_labeldict(self) -> list:
        """Construct the corresponding labels for each bag."""
        raise NotImplementedError

    @abc.abstractmethod
    def read_timeline(self, edgearray) -> list:
        """Read from td_node_status and the edearray to
            - create the timeline of the solving process
            - construct the path and solution-tables used during solving.
        """
        raise NotImplementedError


class DpdbSharpSatVisu(IDpdbVisuConstruct):
    """Implementation of the JSON-construction for the SharpSat problem."""

    def __init__(self, db: pg.extensions.connection,
                 problem: int, intermed_nodes: bool):
        """db : psycopg2.connection
            database to read from.
        problem : int
            index of the problem.
        intermed_nodes : bool
            if True calculates the shortest path between successive nodes.
        """
        LOGGER.debug("Creating %s for problem %d.",
                     self.__class__.__name__, problem)
        self.problem = problem
        self.intermed_nodes = intermed_nodes
        self.num_vars = None

        # wait for good connection
        status = db.get_transaction_status()
        sleeptimer = 0.5
        while status not in good_db_status():
            logging.warning("Waiting %fs for DB connection in status %s",
                            sleeptimer, PSYCOPG2_8_5_TASTATUS[status])
            sleep(sleeptimer)
            status = db.get_transaction_status()

        self.connection = db

    def construct(self) -> dict:
        """
        Construct the Json calling several helper methods.

        Returns
        -------
        dict
            The Json for the visualization-API.

        """

        clauses_edges = self.read_clauses()
        incidence_graph = {
            "var_name_one": 'c_',
            "var_name_two": 'v_',
            "infer_primal": True,
            "edges": clauses_edges}

        # create tree_dec_json
        labeldict = self.read_labeldict()
        edgearray = self.read_edgearray()
        tree_dec_json = {
            "bagpre": "bag %s",
            "edgearray": edgearray,
            "labeldict": labeldict,
            "num_vars": self.read_num_vars()}

        timeline = self.read_timeline(edgearray)

        return {'incidenceGraph': incidence_graph,
                'tdTimeline': timeline,
                'treeDecJson': tree_dec_json}

    def read_num_vars(self) -> int:
        """
        Select the number of vertices in the graph.

        Returns
        -------
        int
            Number of vertices in the graph.

        """
        with self.connection.cursor() as cur:  # create a cursor
            cur.execute(
                "SELECT num_vertices FROM "
                "public.problem WHERE id=%s", (self.problem,))
            self.num_vars = cur.fetchone()[0]
            assert isinstance(self.num_vars, int)
            return self.num_vars

    def read_clauses(self) -> list:
        """Return the clauses used for satisfiability.
        Variables are counted from 1 and negative if negated in the clause.
        For example:

            [{
                "id" : 1,
                "list" : [ 1, -4, 6 ]
            },...]
        """
        with self.connection.cursor() as cur:  # create a cursor
            try:
                cur.execute(
                    f"SELECT * FROM public.p{self.problem}_sat_clause")
            except pg.ProgrammingError:
                LOGGER.error(
                    "dpdb.py SHARPSAT NEEDS TO BE RUN WITH '--store-formula'!")
                raise
            result = cur.fetchall()
            result_cleaned = [[pos if elem else -pos for pos, elem in
                               enumerate(line, 1) if elem is not None]
                              for line in result]
            clauses_edges = [{'id': i, 'list': item}
                             for (i, item) in enumerate(result_cleaned, 1)]
            return clauses_edges

    def read_labeldict(self) -> list:
        """
        Read edges from '_td_bag' and the labels from 'td_node_status' for the bags.

        Returns
        -------
        list
            The filled labeldict for visualization.

        """
        with self.connection.cursor() as cur:  # create a cursor
            labeldict = []
            # check bag numbering:
            cur.execute(
                f"SELECT bag FROM public.p{self.problem}_td_bag group by bag")
            bags = sorted(list(flatten(cur.fetchall())))
            LOGGER.debug("bags: %s", bags)
            for bag in bags:
                cur.execute(
                    f"SELECT node FROM public.p{self.problem}_td_bag WHERE bag=%s", (bag,))
                nodes = list(flatten(cur.fetchall()))
                cur.execute(
                    "SELECT start_time,end_time-start_time "
                    f"FROM public.p{self.problem}_td_node_status WHERE node=%s", (bag,))
                start_time, dtime = cur.fetchone()
                labeldict.append(
                    {'id': bag, 'items': nodes, 'labels':
                     [str(nodes),
                      "dtime=%.4fs" % dtime.total_seconds(),
                      # start_time.strftime("%D %T")
                      ]})
            return labeldict

    def read_timeline(self, edgearray) -> list:
        """
        Read from td_node_status and the edearray to
        - create the timeline of the solving process
        - construct the path and solution-tables used during solving.

        Parameters
        ----------
        edgearray : array of pairs of bagids
            Representing the tree-like structure between all bag-ids.
            It is assumed that all ids are included in this array.
            Example: [(2, 1), (3, 2), (4, 2), (5, 4)]

        Returns
        -------
        result : array
            array of bagids and eventually solution-tables.

        """
        with self.connection.cursor() as cur:  # create a cursor
            timeline = list()
            adj = convert_to_adj(edgearray) if self.intermed_nodes else {}
            cur.execute(
                f"SELECT node FROM public.p{self.problem}_td_node_status ORDER BY start_time")
            order_solved = list(flatten(cur.fetchall()))
            # tour sol -> through result nodes along the edges

            if self.intermed_nodes:
                last = order_solved[-1]
                startpath = find_path(adj, last, order_solved[0])
                timeline = [[bag] for bag in startpath[1]]
            else:
                timeline.append([order_solved[0]])
            # add the other bags in order_solved to the timeline
            last = order_solved[0]
            for bag in order_solved:
                if self.intermed_nodes:
                    path = find_path(adj, last, bag)
                    for intermed in path[1][1:]:
                        timeline.append([intermed])
                # query column names
                #  deepcode ignore Sqli: general query, inserting integers
                cur.execute(
                    "SELECT column_name FROM INFORMATION_SCHEMA.COLUMNS "
                    f"WHERE TABLE_NAME = 'p{self.problem}_td_node_{bag}'")
                column_names = list(flatten(cur.fetchall()))
                LOGGER.debug("column_names %s", column_names)
                # get solutions
                #  deepcode ignore Sqli: general query, inserting integers
                cur.execute(
                    f"SELECT * FROM public.p{self.problem}_td_node_{bag}")
                solution_raw = cur.fetchall()
                LOGGER.debug("solution_raw %s", solution_raw)
                # check for nulled variables - assuming whole columns are
                # nulled:
                columns_notnull = [column_names[i] for i, x in
                                   enumerate(solution_raw[0]) if x is not None]
                solution = [bag,
                            [[columns_notnull,
                              *[[int(v) for v in row if v is not None]
                                for row in solution_raw]],
                             "sol bag " + str(bag),
                             self.footer(solution_raw),
                             True]]
                timeline.append(solution)
                last = bag
            return timeline

    def read_edgearray(self):
        """Read from _td_edge the edges between bags."""
        with self.connection.cursor() as cur:  # create a cursor
            cur.execute(
                f"SELECT node,parent FROM public.p{self.problem}_td_edge")
            result = cur.fetchall()
            return result

    @staticmethod
    def footer(lines) -> str:
        """Returns the footer for solution bags."""
        return "sum: " + str(sum([li[-1] for li in lines]))


class DpdbSatVisu(DpdbSharpSatVisu):
    """Implementation of the JSON-construction for the SAT problem.
    Removing the solution sum in bottom-label.
    """
    @staticmethod
    def footer(lines) -> str:
        """Returns empty footer."""
        return ""


class DpdbMinVcVisu(DpdbSharpSatVisu):
    """Implementation of the JSON-construction for the MinVC problem."""

    def __init__(self, db, problem, intermed_nodes, tw_file=None):
        super().__init__(db, problem, intermed_nodes)
        self.tw_file = tw_file

    def read_clauses(self):
        raise NotImplementedError(
            self.__class__.__name__ +
            " can not read_clauses!")

    @staticmethod
    def footer(lines) -> str:
        """Returns the footer for solution bags."""
        return "min-size: " + str(min([li[-1] for li in lines]))

    def read_twfile(self) -> list:
        """
        Use TwReader.from_file to read the edges for the generalGraph.

        Returns
        -------
        List
            The edges as an list of pairs of vertices.

        """
        LOGGER.info("Reading from %s", self.tw_file)
        try:
            reader = TwReader.from_filewrapper(self.tw_file)
        except Exception as error:
            LOGGER.error("Problem while reading from self.tw_file: %s", error)
            raise error

        # create list so that it is JSON serializable
        return list(reader.edges)

    def construct(self) -> dict:
        """
        Construct the Json calling several helper methods.

        Returns
        -------
        dict
            The Json for the visualization-API.

        """
        # create tree_dec_json
        labeldict = self.read_labeldict()
        edgearray = self.read_edgearray()
        tree_dec_json = {
            'bagpre': "bag %s",
            'edgearray': edgearray,
            'labeldict': labeldict,
            'num_vars': self.read_num_vars()}

        general_gr = {'edges': self.read_twfile()} if self.tw_file else False

        timeline = self.read_timeline(edgearray)
        return {'generalGraph': general_gr,
                'tdTimeline': timeline,
                'treeDecJson': tree_dec_json}


def connect() -> pg.extensions.connection:
    """Connect to the PostgreSQL database server using the params from config."""

    conn = None
    try:
        # read connection parameters
        params = db_config(filename='Archive/database.json')
        db_name = params['database']
        LOGGER.info("Connecting to the PostgreSQL database '%s'...", db_name)
        conn = pg.connect(**params)
        with conn.cursor() as cur:  # create a cursor
            # display the PostgreSQL database server version
            LOGGER.info('PostgreSQL database version:')
            cur.execute('SELECT version()')
            db_version = cur.fetchone()
            LOGGER.info(db_version)
    except (Exception, pg.DatabaseError) as error:
        LOGGER.error(error)
        raise error
    return conn


def create_json(
        problem: int,
        tw_file=None,
        intermed_nodes: bool = False) -> dict:
    """Create the JSON for the specified problem instance."""

    try:
        with connect() as connection:
            # get type of problem
            with connection.cursor() as cur:
                cur.execute("SELECT name,type,num_bags FROM "
                            "public.problem WHERE id=%s", (problem,))
                (name, ptype, num_bags) = cur.fetchone()

            # select the valid constructor for the problem
            constructor: IDpdbVisuConstruct

            if ptype == 'Sat':
                constructor = DpdbSatVisu(
                    connection, problem, intermed_nodes)
            elif ptype == 'SharpSat':
                constructor = DpdbSharpSatVisu(
                    connection, problem, intermed_nodes)
            elif ptype == 'VertexCover':
                constructor = DpdbMinVcVisu(
                    connection, problem, intermed_nodes, tw_file)

            LOGGER.info("Using %s for type=%s",
                        constructor.__class__.__name__, ptype)
            LOGGER.info("Constructing Json...")
            return constructor.construct()
    except (Exception, pg.DatabaseError) as error:
        LOGGER.error(error)
        raise error
    return {}


def main(args: argparse.Namespace) -> None:
    """
    Main method running construct_dpdb_visu for arguments in 'args'

    Parameters
    ----------
    args : argparse.Namespace
        The namespace containing all (command-line) parameters.

    Returns
    -------
    None
    """

    logging_cfg(filename='logging.yml', loglevel=args.loglevel)
    LOGGER.info("Called with '%s'", args)

    problem_ = args.problemnumber
    # get twfile if supplied
    try:
        tw_file_ = args.twfile
    except AttributeError:
        tw_file_ = None
    result_json = create_json(problem=problem_, tw_file=tw_file_)
    # build json filename, can be supplied with problem-number
    try:
        outfile = args.outfile % problem_
    except TypeError:
        outfile = args.outfile
    LOGGER.info("Output file-name: %s", outfile)
    with open(outfile, 'w') as file:
        json.dump(
            result_json,
            file,
            sort_keys=True,
            indent=2 if args.pretty else None,
            ensure_ascii=False)
        LOGGER.debug("Wrote to %s", file)


if __name__ == "__main__":
    # Parse args, call main

    PARSER = argparse.ArgumentParser(
        description="""
        Copyright (C) 2020 Martin Röbke
        This program comes with ABSOLUTELY NO WARRANTY
        This is free software, and you are welcome to redistribute it
        under certain conditions; see COPYING for more information.

        Extracts Information from https://github.com/hmarkus/dp_on_dbs runs
        for further visualization.""",
        epilog=LOGLEVEL_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    PARSER.add_argument('problemnumber', type=int,
                        help="selected problem-id in the postgres-database.")
    PARSER.add_argument('--twfile',
                        type=argparse.FileType('r', encoding='UTF-8'),
                        help="tw-file containing the edges of the graph - "
                        "obtained from dpdb with option --gr-file GR_FILE.")
    PARSER.add_argument('--loglevel', help="set the minimal loglevel for root")
    PARSER.add_argument('--outfile', default='dbjson%d.json',
                        help="default:'dbjson%%d.json'")
    PARSER.add_argument('--pretty', action='store_true',
                        help="pretty-print the JSON.")
    PARSER.add_argument('--inter-nodes', action='store_true',
                        help="calculate and animate the shortest path between "
                        "successive bags in the order of evaluation.")
    PARSER.add_argument('--version', action='version',
                        version='%(prog)s ' + version + ', ' + __date__)

    # get cmd-arguments
    _args = PARSER.parse_args()
    main(_args)



# -*- coding: utf-8 -*-
"""
Visualization for dynamic programming on tree decompositions.

See also the repositories
https://github.com/hmarkus/dp_on_dbs
and
https://github.com/VaeterchenFrost/GPUSAT


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

import argparse
import io
import itertools
import json
import logging
from dataclasses import asdict
from pathlib import Path
from sys import stdin
from typing import Iterable, List, Optional, Union, NewType

from graphviz import Digraph, Graph
from tdvisu.visualization_data import (VisualizationData, IncidenceGraphData,
                                       GeneralGraphData, SvgJoinData)
from tdvisu.version import __date__, __version__ as version
from tdvisu.svgjoin import svg_join
from tdvisu.utilities import flatten, LOGLEVEL_EPILOG, logging_cfg
from tdvisu.utilities import bag_node, solution_node, base_style
from tdvisu.utilities import style_hide_edge, style_hide_node, emphasise_node


LOGGER = logging.getLogger('visualization.py')

StrOrIo = NewType('StrOrIo', Union[str, io.TextIOWrapper])


def read_json(json_data: StrOrIo) -> dict:
    """
    Read json data into a callable object.
    Throws AssertionError if the parsed object has length 0.

    Parameters
    ----------
    json_data : String or io.TextIOWrapper
        The object to be read from.

    Returns
    -------
    result : JSON
        The parsed json.

    """
    if isinstance(json_data, str):
        result = json.loads(json_data)
    elif isinstance(json_data, io.TextIOWrapper):
        result = json.load(json_data)
    else:
        LOGGER.warning("read_json called on %s", type(json_data))
        result = json_data
    assert len(result) > 0, "Please input a valid JSON resource!"
    return result


class Visualization:
    """Holds and processes the information needed to provide dot-format
    and image output for the visualization
    of dynamic programming on tree decomposition.
    """

    def __init__(self, infile: StrOrIo, outfolder: Path):
        """Copy needed fields from arguments and create VisualizationData."""
        self.data: VisualizationData = self.inspect_json(infile)
        self.outfolder = Path(outfolder).resolve()
        self.tree_dec_digraph = None
        LOGGER.debug("Initialized: %s", self)

    def inspect_json(self, infile: StrOrIo) -> VisualizationData:
        """Read and preprocess the needed data from the infile into VisualizationData."""
        LOGGER.debug("Reading from: %s", infile)
        visudata = read_json(infile)
        LOGGER.debug("Found keys: %s", visudata.keys())

        try:
            _incid = visudata.get('incidenceGraph', None)
            _general_graph = visudata.get('generalGraph', None)
            _svg_join = visudata.get('svgjoin', None)

            incid_data: Optional[IncidenceGraphData] = None
            if _incid:
                _incid['edges'] = [[x['id'], x['list']]
                                   for x in _incid['edges']]
                incid_data = IncidenceGraphData(**_incid)
            visudata.pop('incidenceGraph')
            general_graph_data: Optional[GeneralGraphData] = None
            if _general_graph:
                general_graph_data = GeneralGraphData(**_general_graph)
            visudata.pop('generalGraph')
            svg_join_data: Optional[SvgJoinData] = None
            if _svg_join:
                svg_join_data = SvgJoinData(**_svg_join)
            if 'svgjoin' in visudata:
                visudata.pop('svgjoin')

            self.timeline = visudata['tdTimeline']
            visudata.pop('tdTimeline')
            self.tree_dec = visudata['treeDecJson']
            self.bagpre = self.tree_dec['bagpre']
            self.joinpre = self.tree_dec.get('joinpre', 'Join %d~%d')
            self.solpre = self.tree_dec.get('solpre', 'sol%d')
            self.soljoinpre = self.tree_dec.get('soljoinpre', 'solJoin%d~%d')
            visudata.pop('treeDecJson')
        except KeyError as err:
            raise KeyError(f"Key {err} not found in the input Json.")
        return VisualizationData(incidence_graph=incid_data,
                                 general_graph=general_graph_data,
                                 svg_join=svg_join_data,
                                 **visudata)

    def setup_tree_dec_graph(
            self,
            rankdir: str = 'BT',
            shape: str = 'box',
            fillcolor: str = 'white',
            style: str = 'rounded,filled',
            margin: str = '0.11,0.01') -> None:
        """Create self.tree_dec_digraph
        strict means not a multigraph - equal edges get merged.

        rankdir sets the direction in which the nodes are built up.
            - normally Bottom-Top or Top-Bottom.
        """
        self.tree_dec_digraph = Digraph(
            'Tree-Decomposition', strict=True,
            graph_attr={'rankdir': rankdir},
            node_attr={
                'shape': shape,
                'fillcolor': fillcolor,
                'style': style,
                'margin': margin})

    def basic_tdg(self) -> None:
        """Create basic bag structure in tree_dec_digraph."""
        for item in self.tree_dec['labeldict']:
            bagname = self.bagpre % str(item['id'])
            self.tree_dec_digraph.node(bagname,
                                       bag_node(bagname, item['labels']))

        self.tree_dec_digraph.edges([(self.bagpre % str(first), self.bagpre % str(
            second)) for (first, second) in self.tree_dec['edgearray']])

    def forward_iterate_tdg(
            self,
            joinpre: str,
            solpre: str,
            soljoinpre: str) -> None:
        """Create the final positions of all nodes with solutions."""
        tdg = self.tree_dec_digraph                 # shorten name

        for i, node in enumerate(self.timeline):    # Create the positions
            if len(node) > 1:
                # solution to be displayed
                id_inv_bags = node[0]
                if isinstance(id_inv_bags, int):
                    last_sol = solpre % id_inv_bags
                    tdg.node(last_sol, solution_node(
                        *(node[1])), shape='record')

                    tdg.edge(self.bagpre % id_inv_bags, last_sol)

                else:                               # joined node with 2 bags
                    suc = self.timeline[i + 1][0]   # get the joined bags

                    LOGGER.debug('joining %s to %s ', node[0], suc)

                    id_inv_bags = tuple(id_inv_bags)
                    last_sol = soljoinpre % id_inv_bags
                    tdg.node(last_sol, solution_node(
                        *(node[1])), shape='record')

                    tdg.edge(joinpre % id_inv_bags, last_sol)
                    # edges
                    for child in id_inv_bags:  # basically "remove" current
                        # TODO check where 2 args are possibly occuring
                        tdg.edge(
                            self.bagpre % child
                            if isinstance(child, int) else joinpre % child,
                            self.bagpre % suc
                            if isinstance(suc, int) else joinpre % suc,
                            style='invis',
                            constraint='false')
                        tdg.edge(self.bagpre % child if isinstance(child, int)
                                 else joinpre % child,
                                 joinpre % id_inv_bags)
                    tdg.edge(joinpre % id_inv_bags, self.bagpre % suc
                             if isinstance(suc, int) else joinpre % suc)

    def backwards_iterate_tdg(self, joinpre: str, solpre: str, soljoinpre: str,
                              view: bool = False) -> None:
        """Cut the single steps back and update emphasis acordingly."""
        tdg = self.tree_dec_digraph     # shorten name
        last_sol = ""
        _filename = self.outfolder / self.data.td_file
        for i, node in enumerate(reversed(self.timeline)):
            id_inv_bags = node[0]
            LOGGER.debug("%s: Reverse traversing on %s", i, id_inv_bags)

            if i > 0:
                # Delete previous emphasis
                prevhead = self.timeline[len(self.timeline) - i][0]
                bag = (
                    self.bagpre %
                    prevhead if isinstance(
                        prevhead,
                        int) else joinpre %
                    tuple(prevhead))
                base_style(tdg, bag)
                if last_sol:
                    style_hide_node(tdg, last_sol)
                    style_hide_edge(tdg, bag, last_sol)
                    last_sol = ""

            if len(node) > 1:
                # solution to be displayed
                if isinstance(id_inv_bags, int):
                    last_sol = solpre % id_inv_bags
                    emphasise_node(tdg, last_sol)
                    tdg.edge(self.bagpre % id_inv_bags, last_sol)
                else:  # joined node with 2 bags
                    id_inv_bags = tuple(id_inv_bags)
                    last_sol = soljoinpre % id_inv_bags
                    emphasise_node(tdg, last_sol)

            emphasise_node(tdg,
                           self.bagpre %
                           id_inv_bags if isinstance(
                               id_inv_bags,
                               int) else joinpre %
                           id_inv_bags)

            tdg.render(
                view=view, format='svg',
                filename=str(_filename) + str(len(self.timeline) - i)
            )

    def tree_dec_timeline(self, view: bool = False) -> None:
        """Main-method for handling all construction of the timeline."""

        self.setup_tree_dec_graph(
            rankdir=self.data.orientation,
            fillcolor=self.data.bagcolor)
        self.basic_tdg()

        # Iterate labeldict
        self.forward_iterate_tdg(
            joinpre=self.joinpre,
            solpre=self.solpre,
            soljoinpre=self.soljoinpre)
        self.backwards_iterate_tdg(
            view=view,
            joinpre=self.joinpre,
            solpre=self.solpre,
            soljoinpre=self.soljoinpre)

        # Prepare supporting graph timeline

        _timeline: List[Optional[List[int]]] = []
        for step in self.timeline:
            if len(step) < 2:
                _timeline.append(None)
            elif isinstance(step[0], int):
                _timeline.append(
                    next(
                        (item.get('items') for item in self.tree_dec['labeldict']
                         if item['id'] == step[0])))
            else:
                # Join operation - no clauses involved in computation
                _timeline.append(None)
        __incid = self.data.incidence_graph
        if __incid:
            if __incid.infer_primal or __incid.infer_dual:
                # prepare incid edges with abs:
                abs_clauses = [[cl[0], list(map(abs, cl[1]))]
                               for cl in __incid.edges]
            if __incid.infer_primal:
                # vertex for each variable + edge if the variables
                # occur in the same clause:
                primal_edges = set(flatten(             # remove duplicates
                    [itertools.combinations(cl[1], 2)
                     for cl in abs_clauses]))
                # check if any node is really isolated:
                isolated = [cl[1][0] for cl in abs_clauses
                            if len(cl[1]) == 1 and
                            not any(cl[1][0] in sl for sl in primal_edges)]

                self.general_graph(
                    timeline=_timeline,
                    edges=primal_edges,
                    extra_nodes=set(isolated),
                    graph_name=__incid.primal_file,
                    file_basename=__incid.primal_file,
                    var_name=__incid.var_name_two)
                LOGGER.info("Created infered primal-graph")

            if __incid.infer_dual:
                # Edge, if clauses share the same variable
                dual_edges = [(cl[0], other[0])
                              for i, cl in enumerate(abs_clauses)
                              for other in abs_clauses[i + 1:]  # no multiples
                              if any(var in cl[1] for var in other[1])]
                # check if any clause is isolated:
                isolated = [cl[0] for cl in abs_clauses
                            if not any(cl[0] in sl for sl in dual_edges)]

                self.general_graph(
                    timeline=_timeline,
                    edges=dual_edges,
                    extra_nodes=set(isolated),
                    graph_name=__incid.dual_file,
                    file_basename=__incid.dual_file,
                    var_name=__incid.var_name_one)
                LOGGER.info("Created infered dual-graph")
            self.incidence(
                timeline=_timeline,
                inc_file=__incid.inc_file,
                num_vars=self.tree_dec['num_vars'],
                colors=self.data.colors, view=view,
                fontsize=__incid.fontsize,
                penwidth=__incid.penwidth,
                basefill=self.data.bagcolor,
                var_name_one=__incid.var_name_one,
                var_name_two=__incid.var_name_two,
                column_distance=__incid.column_distance)
            LOGGER.info(
                "Created incidence-graph for file='%s'",
                self.data.incidence_graph.inc_file)
        if self.data.general_graph:
            self.general_graph(timeline=_timeline, view=view,
                               **asdict(self.data.general_graph))
            LOGGER.info(
                "Created general-graph for file='%s'",
                self.data.general_graph.file_basename)
        if self.data.svg_join:
            self.call_svgjoin()

    def general_graph(
            self,
            timeline: Iterable[Optional[List[int]]],
            edges: Iterable[Iterable[int]],
            extra_nodes: Iterable[int] = tuple(),
            view: bool = False,
            fontsize: int = 20,
            fontcolor: str = 'black',
            penwidth: float = 2.2,
            first_color: str = 'yellow',
            first_style: str = 'filled',
            second_color: str = 'green',
            second_style: str = 'dotted,filled',
            third_color: str = 'red',
            graph_name: str = 'graph',
            file_basename: str = 'graph',
            do_sort_nodes: bool = True,
            do_adj_nodes: bool = True,
            var_name: str = '') -> None:
        """
        Creates one graph emphasized for the given timeline.

        Parameters
        ----------
        edges : Iterable of: {int, int}
            All edges between nodes in the graph.
            Should NOT contain self-edges!
            BOTH edges (x, y) and (y, x) could be in the edgelist.

        extra_nodes : Iterable of int
            Nodes that are probably not in the edges, but should be rendered.

        TIMELINE : Iterable of: None | [int...]
            None if no variables get highlighted in this step.
            Else the 'timeline' provides the set of variables that are
            in the bag(s) under consideration. This function computes all other
            variables that are involved in this timestep using the 'edgelist'.

        colors : Iterable of color
            Colors to use for the graph parts.

        Returns
        -------
        None, but outputs the files with the graph for each timestep.

        """
        _filename = self.outfolder / file_basename
        LOGGER.info("Generating general-graph for '%s'", file_basename)
        vartag_n: str = var_name + '%d'
        # sfdp http://yifanhu.net/SOFTWARE/SFDP/index.html
        default_engine = 'sfdp'

        graph = Graph(
            graph_name,
            strict=True,
            engine=default_engine,
            graph_attr={
                'fontsize': str(fontsize),
                'overlap': 'false',
                'outputorder': 'edgesfirst',
                'K': '2'},
            node_attr={
                'fontcolor': str(fontcolor),
                'penwidth': str(penwidth),
                'style': 'filled',
                'fillcolor': 'white'})

        if do_sort_nodes:
            bodybaselen = len(graph.body)
            # 1: layout with circo
            graph.engine = 'circo'
            # 2: nodes in edges+extra_nodes make a circle
            nodes = sorted([vartag_n % n for n in set(
                itertools.chain(flatten(edges), extra_nodes))])
            for i, node in enumerate(nodes):
                graph.edge(str(nodes[i - 1]), str(node))
            # 3: reads in bytes!
            code_lines = graph.pipe('plain').splitlines()
            # 4: save the (sorted) positions
            assert code_lines[0].startswith(b'graph')
            node_positions = [line.split()[1:4] for line in code_lines[1:]
                              if line.startswith(b'node')]
            # 5: cut layout
            graph.body = graph.body[:bodybaselen]
            for line in node_positions:
                graph.node(line[0].decode(),
                           pos='%f,%f!' % (float(line[1]), float(line[2])))
            # 6: Engine uses previous positions
            graph.engine = 'neato'

        for (src, tar) in edges:
            graph.edge(vartag_n % src, vartag_n % tar)
        for nodeid in extra_nodes:
            graph.node(vartag_n % nodeid)

        bodybaselen = len(graph.body)

        for i, variables in enumerate(timeline, start=1):    # all timesteps
            # reset highlighting
            graph.body = graph.body[:bodybaselen]

            if variables is None:
                graph.render(
                    view=view,
                    format='svg',
                    filename=str(_filename) + str(i))
                continue

            for var in variables:
                graph.node(
                    vartag_n % var,
                    fillcolor=first_color,
                    style=first_style)

            # highlight edges between variables
            for (s, t) in edges:
                if s in variables and t in variables:
                    graph.edge(
                        vartag_n % s,
                        vartag_n % t,
                        color=third_color,
                        penwidth=str(penwidth))

            if do_adj_nodes:
                # set.difference accepts list as argument, "-" does not.
                edges = [set(edge) for edge in edges]
                adjacent = {
                    edge.difference(variables).pop() for edge in edges if len(
                        edge.difference(variables)) == 1}

                for var in adjacent:
                    graph.node(vartag_n % var,
                               color=second_color,
                               style=second_style)

            graph.render(view=view, format='svg',
                         filename=str(_filename) + str(i))

    def incidence(
            self,
            timeline: Iterable[Optional[List[int]]],
            num_vars: int,
            colors: List,
            inc_file: str = 'IncidenceGraphStep',
            view: bool = False,
            fontsize: Union[str, int] = 16,
            penwidth: float = 2.2,
            basefill: str = 'white',
            sndshape: str = 'diamond',
            neg_tail: str = 'odot',
            var_name_one: str = '',
            var_name_two: str = '',
            column_distance: float = 0.5) -> None:
        """
        Creates the incidence graph emphasized for the given timeline.

        Parameters
        ----------
        timeline : Iterable of: None | [int...]
            None if no variables get highlighted in this step.
            Else the 'timeline' provides the set of variables that are
            in the bag(s) under consideration. This function computes all other
            variables that are involved in this timestep using the 'edgelist'.
        num_vars : int
            Count of variables that are used in the clauses.
        colors : Iterable of color
            Colors to use for the graph parts.
        inc_file : string
            Basis for the file created, gets appended with the step number.
        view : bool
            If true opens the created file after creation. Default is false.
        basefill : color
            Background color of all nodes. Default is 'white'.
        sndshape : string
            Description of the shape for nodes with the variables. Default diamond.
        neg_tail : string
            Description of the shape of the edge-tail indicating a
            negated variable. Default is 'odot'.
        column_distance : float
            Changes the distance between both partitions, measured in image-units.
            Default is 0.5

        Returns
        -------
        None, but outputs the files with the graph for each timestep.

        """
        _filename = self.outfolder / inc_file

        clausetag_n = var_name_one + '%d'
        vartag_n = var_name_two + '%d'

        g_incid = Graph(
            inc_file,
            strict=True,
            graph_attr={
                'splines': 'false',
                'ranksep': '0.2',
                'nodesep': str(float(column_distance)),
                'fontsize': str(
                    int(fontsize)),
                'compound': 'true'},
            edge_attr={
                'penwidth': str(float(penwidth)),
                'dir': 'back',
                'arrowtail': 'none'})
        __incid = self.data.incidence_graph
        with g_incid.subgraph(name='cluster_clause',
                              edge_attr={'style': 'invis'},
                              node_attr={'style': 'rounded,filled',
                                         'fillcolor': basefill}) as clauses:
            clauses.attr(label='clauses')
            clauses.edges([(clausetag_n % (i + 1), clausetag_n % (i + 2))
                           for i in range(len(__incid.edges) - 1)])

        g_incid.attr('node', shape=sndshape,
                     penwidth=str(float(penwidth)),
                     style='dotted')
        with g_incid.subgraph(name='cluster_ivar',
                              edge_attr={'style': 'invis'}) as ivars:
            ivars.attr(label='variables')
            ivars.edges([(vartag_n % (i + 1), vartag_n % (i + 2))
                         for i in range(num_vars - 1)])
            for i in range(num_vars):
                g_incid.node(vartag_n %
                             (i + 1), vartag_n %
                             (i + 1), color=colors[(i + 1) %
                                                   len(colors)])

        g_incid.attr('edge', constraint='false')

        for clause in __incid.edges:
            for var in clause[1]:
                if var >= 0:
                    g_incid.edge(clausetag_n % clause[0],
                                 vartag_n % var,
                                 color=colors[var % len(colors)])
                else:
                    g_incid.edge(clausetag_n % clause[0],
                                 vartag_n % -var,
                                 color=colors[-var % len(colors)],
                                 arrowtail=neg_tail)

        # make edgelist variable-based (varX, clauseY), ...
        #  var_cl_iter [(1, 1), (4, 1), ...
        var_cl_iter = tuple(flatten([[(x, y[0]) for x in y[1]]
                                     for y in __incid.edges]))

        bodybaselen = len(g_incid.body)
        for i, variables in enumerate(timeline, start=1):    # all timesteps

            # reset highlighting
            g_incid.body = g_incid.body[:bodybaselen]
            if variables is None:
                g_incid.render(view=view, format='svg',
                               filename=str(_filename) + str(i))
                continue

            emp_clause = {
                var_cl[1] for var_cl in var_cl_iter if abs(
                    var_cl[0]) in variables}

            emp_var = {abs(var_cl[0])
                       for var_cl in var_cl_iter if var_cl[1] in emp_clause}

            for var in emp_var:
                _vartag = vartag_n % abs(var)
                _style = 'solid,filled' if var in variables else 'dotted,filled'
                g_incid.node(
                    _vartag,
                    _vartag,
                    style=_style,
                    fillcolor='yellow')

            for clause in emp_clause:
                g_incid.node(
                    clausetag_n % clause,
                    clausetag_n % clause,
                    fillcolor='yellow')

            for edge in var_cl_iter:
                (var, clause) = edge

                _style = 'solid' if clause in emp_clause else 'dotted'
                _vartag = vartag_n % abs(var)

                if var >= 0:
                    g_incid.edge(clausetag_n % clause,
                                 _vartag,
                                 color=colors[var % len(colors)],
                                 style=_style)
                else:                                       # negated variable
                    g_incid.edge(clausetag_n % clause,
                                 _vartag,
                                 color=colors[-var % len(colors)],
                                 arrowtail='odot',
                                 style=_style)

            g_incid.render(
                view=view,
                format='svg',
                filename=str(_filename) + str(i))

    def call_svgjoin(self) -> None:
        """Analyzes content in data.svg_join for the call to svg_join."""
        sj_data = self.data.svg_join
        if not sj_data.base_names:
            LOGGER.warning(
                "svg_join data in JsonAPI contains no file-names to join.")
            return
        if isinstance(sj_data.base_names, str):
            sj_data.base_names = [sj_data.base_names]
        sj_data.num_images = int(sj_data.num_images)
        # Other arguments get handled directly in svgjoin for iterators etc.
        # Use default outfolder only if folder is None
        if sj_data.folder is None:
            updated_sj_data = asdict(sj_data)
            updated_sj_data["folder"] = self.outfolder
            svg_join(**updated_sj_data)
        else:
            svg_join(**asdict(sj_data))


def main(args: argparse.Namespace) -> None:
    """
    Main method running construct_dpdb_visu for arguments in 'args'

    Parameters
    ----------
    args : argparse.Namespace
        The namespace containing all (command-line) parameters.

    Returns
    -------
    None
    """

    logging_cfg(filename='logging.yml', loglevel=args.loglevel)
    LOGGER.info("Called with '%s'", args)

    infile = args.infile
    outfolder = args.outfolder
    if not outfolder:
        outfolder = 'outfolder'
    outfolder = Path(outfolder).resolve()

    LOGGER.info("Will read from '%s' and write to folder '%s'",
                infile.name, outfolder)
    visu = Visualization(infile=infile, outfolder=outfolder)
    visu.tree_dec_timeline()


if __name__ == "__main__":
    # Parse args, call main

    PARSER = argparse.ArgumentParser(
        description="""
        Copyright (C) 2020 Martin Röbke
        This program comes with ABSOLUTELY NO WARRANTY
        This is free software, and you are welcome to redistribute it
        under certain conditions; see COPYING for more information.

        Visualizing Dynamic Programming on Tree-Decompositions.""",
        epilog=LOGLEVEL_EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    # possible to use stdin for the file.
    PARSER.add_argument('infile', nargs='?',
                        type=argparse.FileType('r', encoding='UTF-8'),
                        default=stdin,
                        help="Input file for the visualization "
                        "must conform with the 'JsonAPI.md'")
    PARSER.add_argument('outfolder',
                        help="Folder to output the visualization results")
    PARSER.add_argument('--version', action='version',
                        version='%(prog)s ' + version + ', ' + __date__)
    PARSER.add_argument('--loglevel', help="set the minimal loglevel for root")

    # get cmd-arguments
    _args = PARSER.parse_args()
    # call main()
    main(_args)



# -*- coding: utf-8 -*-
"""
Adapted DIMACS reader from: https://github.com/hmarkus/dp_on_dbs/tree/master/dpdb
DIMACS is an acronym derived from http://dimacs.rutgers.edu/
Discrete Mathematics and Theoretical Computer Science.

Example usage:

import from tdvisu.reader import TwReader

def read_tw_file(tw_file)
    reader = TwReader()
    inp = r.from_filename(tw_file)

def main()
    file = 'gr5.tw'
    graph = read_tw_file(file)
    print(graph.num_vertices)
    print(graph.num_edges)
    print(graph.edges)


Copyright (C) 2020  Patrick Thier https://github.com/Fend0r, TU Wien,
                    Martin Röbke, TU Dresden

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""
from __future__ import annotations
import logging
import sys
from typing import Iterable

from tdvisu.utilities import add_edge_to

logger = logging.getLogger('reader.py')


class Reader():
    """Base class for string-readers."""
    @classmethod
    def from_filename(cls, fname) -> Reader:
        with open(fname, "r") as file:
            return cls.from_string(file.read())

    @classmethod
    def from_filewrapper(cls, fwrapper) -> Reader:
        return cls.from_string(fwrapper.read())

    @classmethod
    def from_stream(cls, stream) -> Reader:
        return cls.from_string(stream.read().decode())

    @classmethod
    def from_string(cls, string) -> Reader:
        instance = cls()
        instance.parse(string)
        return instance

    def parse(self, string):
        pass


class DimacsReader(Reader):
    """Reader for the DIMACS graph data format.
    It is a commonly used exchange format for graphs.
    It stores a single undirected graph.
    Plain text or binary format.
    DIMACS is an acronym derived from http://dimacs.rutgers.edu/
    Discrete Mathematics and Theoretical Computer Science.
    """

    def __init__(self):
        self.problem_solution_type = None
        self.format = None
        self._problem_vars = None

    def parse(self, string) -> None:
        lines = string.split("\n")
        body_start = self.preamble(lines)
        self.store_problem_vars()
        self.body(lines[body_start:])

    def store_problem_vars(self):
        pass

    @staticmethod
    def is_comment(line: str) -> bool:
        return line.startswith("c ") or line == "c"

    def body(self, lines):
        pass

    def preamble(self, lines: Iterable[str]) -> int:
        """
        Searches for the preamble line in lines and saves:
            problem_solution_type, format, _problem_vars
        Then returns the index of the next line.

        If no preamble is present:
            Exit with error.

        """
        for lineno, line in enumerate(lines):
            if line.startswith("p ") or line.startswith("s "):
                line = line.split()
                self.problem_solution_type = line[0]
                self.format = line[1]
                self._problem_vars = line[2:]
                return lineno + 1

            if not line or self.is_comment(line):
                continue

            warn = "Invalid content in preamble at line %d: %s"
            logger.warning(warn, lineno, line)

        logger.error("No type found in DIMACS file!")
        sys.exit(1)


class TwReader(DimacsReader):
    """Dimacs Reader for the 'tw' format (saving directed edges).
    Stores edges and adjacency together with
    the number of vertices and number of edges.
    """

    def __init__(self):
        super().__init__()
        self.edges = set()
        self.adjacency_dict = dict()
        self.num_vertices = self.num_edges = None

    def store_problem_vars(self):
        self.num_vertices = int(self._problem_vars[0])
        self.num_edges = int(self._problem_vars[1])

    def body(self, lines) -> None:
        """Store the content from the given lines in the edges and adjacency_dict."""
        if self.format not in ('col', 'tw'):
            logger.error("Not a tw file!")
            sys.exit(1)

        for lineno, line in enumerate(lines):
            if not line or self.is_comment(line):
                continue

            line = line.split()
            if len(line) != 2:
                logger.warning(
                    "Expected exactly 2 vertices at line %d, but %d found",
                    lineno,
                    len(line))
            vertex1 = int(line[0])
            vertex2 = int(line[1])

            add_edge_to(self.edges, self.adjacency_dict, vertex1, vertex2)

        if len(self.edges) != self.num_edges:
            logger.warning(
                "Number of edges mismatch preamble (%d vs %d)", len(
                    self.edges), self.num_edges)



# -*- coding: utf-8 -*-
"""
Testing dijkstra.py


Copyright (C) 2020  Martin Röbke

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.
    If not, see https://www.gnu.org/licenses/gpl-3.0.html

"""

from pytest import mark, raises
from tdvisu.dijkstra import DijkstraNoPath, bidirectional_dijkstra as find_path


class TestShortestPath:
    """
    Test the functionality of find_path for simple graphs.
    Should find the shortest path on undirected edges if one is availabe,
    raise ValueError if the source or target are not in the edges,
    and raise DijkstraNoPath otherwise.
    """
    edges_low = {2: {1: {}, 3: {}, 4: {}},
                 1: {2: {}},
                 3: {2: {}},
                 4: {2: {}, 5: {}},
                 5: {4: {}}}

    edges_high = {20: {10: {}, 30: {}, 40: {}},
                  10: {20: {}},
                  30: {20: {}},
                  40: {20: {}, 50: {}},
                  50: {40: {}}}

    def test_normal_path(self):
        """Can find a short path."""
        result = find_path(self.edges_low, 1, 4)
        assert result == (2, [1, 2, 4])

    @mark.parametrize(
        "arg",
        [{1: {}}, edges_low]
    )
    def test_length0(self, arg):
        """Staying on one node."""
        result = find_path(arg, 1, 1)
        assert result == (0, [1])

    @mark.parametrize(
        "source,target",
        [(1, 1), (2, 2), ('a', 'a'), ('33', '23'), (100000, 2000000)]
    )
    def test_empty_graph(self, source, target):
        """Empty graph should raise Value Error."""
        with raises(ValueError):
            find_path({}, source, target)

    def test_source_not_in_graph(self):
        """Should raise ValueError"""
        with raises(ValueError):
            find_path(self.edges_low, -1, 4)

    def test_target_not_in_graph(self):
        """Should raise ValueError"""
        with raises(ValueError):
            find_path(self.edges_low, 1, -4)

    def test_no_path(self):
        """Should raise ValueError"""
        with raises(DijkstraNoPath):
            find_path({**self.edges_low, **self.edges_high}, 1, 10)

    @mark.parametrize(
        "weight",
        [2, 100, 0.25, 0.1, 0.15]
    )
    def test_simple_weight(self, weight):
        """Weight is one constant function"""
        result = find_path(self.edges_low, 1, 4, lambda u, v, data: weight)
        assert result == (2 * weight, [1, 2, 4])

    @mark.parametrize(
        "weight",
        [2, 100, 0.1, 0.15]
    )
    def test_individual_weight(self, weight):
        """Weight is attribute in edges"""
        edges = {k: {target: {'weight': weight}for target in d}
                 for k, d in self.edges_low.items()}
        result = find_path(edges, 1, 4)
        assert result == (2 * weight, [1, 2, 4])



# -*- coding: utf-8 -*-
"""Describing local installs or distribution of the package tdvisu."""

from setuptools import setup
from tdvisu.version import __version__ as version


def read_files(files, delim: str = "\n") -> str:
    r"""
    Concatenate the content of one or more files joined by a delimiter.

    Parameters
    ----------
    files : Iterable of single files
        Every file is a text or byte string giving the name
        (and the path if the file isn't in the current working directory)
         of the file to be opened or an integer file descriptor
         of the file to be wrapped.
    delim : str, optional
        The delimiter to be inserted between the contents. The default is "\n".

    Returns
    -------
    str
        The concatenated string.
    """
    data = []
    try:
        for file in files:
            with open(file, encoding='utf-8') as handle:
                data.append(handle.read())
    except IOError:
        pass
    return delim.join(data)


long_description = read_files(['README.md', 'CHANGELOG.md'])


description = "Visualizing Dynamic Programming on Tree Decompositions."

classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Education',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Topic :: Scientific/Engineering :: Visualization',
    'Topic :: Multimedia :: Graphics :: Presentation']

tests_require = ['pytest']

setup(name="tdvisu",
      version=version,
      description=description,
      long_description=long_description,
      long_description_content_type='text/markdown',
      url="https://github.com/VaeterchenFrost/tdvisu",
      author="Martin Röbke",
      author_email="martin.roebke@mailbox.tu-dresden.de",
      license='GPLv3',
      packages=['tdvisu'],
      platforms='any',
      install_requires=['graphviz', 'psycopg2', 'python-benedict'],
      extras_require={'test': tests_require},
      classifiers=classifiers,
      keywords='graph visualization dynamic-programming msol-solver')
